package de.tudresden.inf.st.ros2rag.starter;

import de.tudresden.inf.st.ros2rag.starter.ast.*;
import org.apache.logging.log4j.LogManager;
import org.apache.logging.log4j.Logger;

import java.io.IOException;
import java.util.concurrent.CountDownLatch;
import java.util.concurrent.TimeUnit;

/**
 * Testing Ros2Rag without generating something.
 *
 * @author rschoene - Initial contribution
 */
public class Main {

  private static final Logger logger = LogManager.getLogger(Main.class);

  public static void main(String[] args) throws IOException, InterruptedException {
    // parse arguments
    final String mqttHost;
    if (args.length > 0) {
      mqttHost = args[0];
    } else {
      mqttHost = "localhost";
    }

    /* assumed configuration:
      panda_ground: ground-plate
      panda_link_0: link_0 of the panda
      panda_link_1: link_1 of the panda
      panda_link_2: link_2 of the panda
      panda_link_3: link_3 of the panda
      panda_link_4: link_4 of the panda
      panda_link_5: link_5 of the panda
      panda_link_6: link_6 of the panda
      panda_link_7: link_7 of the panda (end effector)
      panda_link_8: link_8 of the panda (left finger)
      panda_link_9: link_9 of the panda (right finger)
     */
    final int numberOfJoints = 10;
    final int endEffectorIndex = 7;
    final String jointTopicFormat = "panda_link_%d";
    final String configTopic = "robotconfig";

    final int[][] safetyZoneCoordinates = {
        { 0, 0, 0},
        {-1, 0, 0},
        { 1, 0, 0}
    };

    // --- No configuration below this line ---

    Model model = new Model();
    model.MqttSetHost(mqttHost);

    ZoneModel zoneModel = new ZoneModel();
    zoneModel.setSize(makePosition(1, 1, 1));

    Zone safetyZone = new Zone();
    for (int[] coordinate : safetyZoneCoordinates) {
      safetyZone.addCoordinate(new Coordinate(
          IntPosition.of(coordinate[0], coordinate[1], coordinate[2])));
    }
    model.setZoneModel(zoneModel);
    model.MqttWaitUntilReady(2, TimeUnit.SECONDS);

    RobotArm robotArm = new RobotArm();
    model.setRobotArm(robotArm);

    for (int jointIndex = 0; jointIndex < numberOfJoints; jointIndex++) {
      final Joint jointOrEndEffector;
      if (jointIndex == endEffectorIndex) {
        EndEffector endEffector = new EndEffector();
        endEffector.setName("gripper");
        robotArm.setEndEffector(endEffector);
        jointOrEndEffector = endEffector;
      } else {
        Joint joint = new Joint();
        joint.setName("joint" + jointIndex);
        robotArm.addJoint(joint);
        jointOrEndEffector = joint;
      }
      jointOrEndEffector.setCurrentPosition(makePosition(0, 0, 0));
      robotArm.addDependency1(jointOrEndEffector);
      jointOrEndEffector.connectCurrentPosition(String.format(jointTopicFormat, jointIndex));
    }

    robotArm.connectAppropriateSpeed(configTopic, true);

    logStatus("Start", robotArm);
    CountDownLatch exitCondition = new CountDownLatch(1);

    logger.info("To print the current model states, send a message to the topic 'model'.");
    logger.info("To exit the system cleanly, send a message to the topic 'exit'.");

    MqttUpdater mainHandler = new MqttUpdater("mainHandler");
    mainHandler.setHost(mqttHost, 1883);
    mainHandler.waitUntilReady(2, TimeUnit.SECONDS);
    mainHandler.newConnection("exit", bytes -> exitCondition.countDown());
    mainHandler.newConnection("model", bytes -> logStatus(new String(bytes), robotArm));

    exitCondition.await();

    mainHandler.close();
    model.MqttCloseConnections();
  }

  private static void logStatus(String prefix, RobotArm robotArm) {
    StringBuilder sb = new StringBuilder(prefix).append("\n")
        .append("robotArm.isInSafetyZone: ").append(robotArm.isInSafetyZone())
        .append(", robotArm.getAppropriateSpeed = ").append(robotArm.getAppropriateSpeed()).append("\n");
    for (Joint joint : robotArm.getJointList()) {
      sb.append(joint.getName()).append(": ").append(joint.getCurrentPosition()).append("\n");
    }
    sb.append("endEffector ").append(robotArm.getEndEffector().getName()).append(": ")
        .append(robotArm.getEndEffector().getCurrentPosition());
    logger.info(sb.toString());
  }

  private static IntPosition makePosition(int x, int y, int z) {
    return IntPosition.of(x, y, z);
  }
}
